module Compile where



import Control.Monad
import Data.Graph
import Data.List
import Data.Map (Map)
import qualified Data.Map as Map

import Core
import DesugarType
import DesugarExpr
import ErrorHandling
import Names
import Pretty
import Scope
import qualified Syntax as S
import Substitute
import TypeInference
import HaskellBackend

import Debug.Trace



compile :: Scope
		-> [S.Decl]
		-> ErrorHandling Scope

compile scope0 decls = do
	
	-- Compute kinds
	types <- checkForDuplicates "type" (Map.keys $ scopeTypes scope0) [let
		info = case S.declCons decl of
			S.UserDefined cons -> AlgebraicTypeInfo {
				typeInfoFreeVars = S.declFreeVars decl,
				typeInfoCons = map fst cons
				}
			S.Builtin impl -> BuiltinTypeInfo {
				typeInfoFreeVars = S.declFreeVars decl,
				typeInfoImplHaskell = Just impl
				}
		in (S.declTypeName decl, S.declSpan decl, info)
		| decl @ S.DataDecl {} <- decls]
	
	let scope1 = scope0 {
		scopeTypes = types `Map.union` scopeTypes scope0
		}
	
	-- Desugar data type declarations
	cons <- checkForDuplicates "constructor" (Map.keys $ scopeCons scope1) =<< sequence [
	
		errorLoc ("In declaration of data type '" ++ pretty typeName ++ "' at " ++
			pretty (S.declSpan decl) ++ ":") $
			
		errorLoc ("In constructor '" ++ pretty conName ++ "':") $ do
		
			fields' <- sequence [
				desugarFieldType (scope1, Map.fromList vars, Nothing) field
				| (field,n) <- zip fields [1..]]
			
			let kind = case Map.lookup typeName types of Just ti -> typeInfoKind ti
			let returnType = foldl ApplyType (ConType typeName kind) [VarType v k | (v,k) <- vars]
			
			return (
				conName,
				S.declSpan decl,
				ConInfo {
					conFreeVars = vars,
					conFieldTypes = fields',
					conReturnType = returnType
				})
				
		| decl @ S.DataDecl {
			S.declFreeVars = vars,
			S.declTypeName = typeName,
			S.declCons = S.UserDefined cons } <- decls
		, (conName, fields) <- cons]
	
	let scope2 = scope1 {
		scopeCons = cons `Map.union` scopeCons scope1
		}
		
	-- Check for orphaned type signatures
	sequence [do
		case [() | S.DefnDecl { S.declName = name' } <- decls, name == name'] of
			[] -> fail ("Type signature at " ++ pretty (S.declSpan decl) ++ " is for '" ++
				pretty name ++ "', but '" ++ pretty name ++ "' isn't defined.")
			_ -> return ()
		| decl @ S.TypeSignatureDecl { S.declName = name } <- decls]
	
	-- Pair type signatures with definitions
	defns <- checkForDuplicates "value" (Map.keys $ scopeValues scope2) =<< sequence [do
		
		-- Find the type signature, if there is one
		qt <- case [(qt, span)
			| S.TypeSignatureDecl { S.declSpan = span, S.declName = name', S.declType = qt } <- decls
			, name == name'] of
		
			-- No type signature
			[] -> return Nothing
			
			-- Type signature; process it
			[(S.PolyFlavorType vars (S.FlavorType f t), span)] -> do
			
				(fk, f') <- errorLoc ("In flavor signature of '"++pretty name++"' at "++pretty span++":") $
					desugarType (scope1, Map.fromList vars, Nothing) f
				verifyKind fk monadKind ("flavor of '"++pretty name++"' in type signature at "++pretty span)
				
				(tk, t') <- errorLoc ("In type signature of '"++pretty name++"' at "++pretty span++":") $
					desugarType (scope1, Map.fromList vars, Just f') t
				verifyKind tk TypeKind ("type of '"++pretty name++"' in type signature at "++pretty span)
				
				return (Just (PolyFlavorType vars (FlavorType f' t')))
			
			-- Multiple type signatures
			sigs -> fail ("There is more than one type signature for '"++pretty name++"':" ++
				concat ["\n" ++ pretty span | (_, span) <- sigs])
		
		return (name, S.declSpan decl, Defn { defnType = qt, defnExpr = S.declExpr decl })
		
		| decl @ S.DefnDecl { S.declName = name } <- decls]
	
	-- Desugaring of expressions and type inference
	values <- compileDefns scope2 defns
	
	let scope3 = scope2 {
		scopeValues = values `Map.union` scopeValues scope2
		}
	
	-- Desugar transformers
	transformers <- checkForDuplicates "transformer" (Map.keys $ scopeTransformers scope3) =<< sequence [
	
		errorLoc ("In definition of transformer '"++pretty mName++"' at "++pretty span++":") $ do
			
			-- Verify that the transformer is correctly kinded by building a syntax expression to
			-- represent the monad and then checking its kind. We pass 'span' as the location of
			-- each part, even though this is the textual span of the entire monad instance rather
			-- than just the part in question, which might make the error messages a bit weird, but
			-- this doesn't matter.
			let transformerAsSyntax = foldl
				(S.ApplyType span)
				(S.ConType span mName)
				[S.VarType span v | (v,_) <- (S.declFreeVars decl)]
			(tk, transformer) <- desugarType
				(scope3, Map.fromList (S.declFreeVars decl), Nothing) transformerAsSyntax
			verifyKind tk transformerKind "transformer"
			
			let f = VarType (Name "_f") monadKind
			let a = VarType (Name "_a") TypeKind
			let b = VarType (Name "_b") TypeKind
			let t m x = transformer `ApplyType` m `ApplyType` x
			let l `fun` r = FunctionType (StrictField l) (FlavorType f r)
			
			let returnType =
				PolyFlavorType
					([(Name "_a", TypeKind), (Name "_f", monadKind)] ++ S.declFreeVars decl)
					(FlavorType f (a `fun` t f a))
			ret' <- errorLoc ("In definition of 'return':") $
				compileExpr scope3 (Defn {
					defnType = Just returnType,
					defnExpr = S.UserDefined (fst (S.declTransformerDefn decl))
					})
			
			let bindType =
				PolyFlavorType
					([(Name "_a", TypeKind), (Name "_b", TypeKind), (Name "_f", monadKind)] ++ S.declFreeVars decl)
					(FlavorType f (t f a `fun` ((a `fun` t f b) `fun` t f b)))
			bind' <- errorLoc ("In definition of 'bind':") $
				compileExpr scope3 (Defn {
					defnType = Just bindType,
					defnExpr = S.UserDefined (snd (S.declTransformerDefn decl))
					})
			
			return (
				mName,
				span,
				TransformerInfo {
					transformerReturn = ret',
					transformerBind = bind'
				})
			| decl @ S.TransformerDecl { S.declSpan = span, S.declTransformerName = mName } <- decls]
	
	let scope4 = scope3 {
		scopeTransformers = transformers `Map.union` scopeTransformers scope3
		}
	
	return scope4

	where
		checkForDuplicates :: String -> [Name] -> [(Name, S.SrcSpan, b)] -> ErrorHandling (Map Name b)
		checkForDuplicates _ _ [] = return Map.empty
		checkForDuplicates what predefined ((name, span, value) : xs) =
			if any (\(name', _, _) -> name == name') xs || name `elem` predefined
				then do
					let dups =
						[pretty span] ++
						[pretty span | (name', span, _) <- xs, name == name'] ++
						(if name `elem` predefined then ["<predefined>"] else [])
					fail ("There is more than one " ++ what ++ " named '" ++ pretty name ++ "':" ++
						concat ["\n" ++ thing | thing <- dups])
				else do
					rest <- checkForDuplicates what predefined xs
					return (Map.insert name value rest)


{-
There are two entry points to the type inference system:

compileDefns compiles a group of definitions, which may be mutually recursive.

compileExpr compiles a single definition.
-}

data Defn = Defn {
	defnType :: Maybe PolyFlavorType,
	defnExpr :: S.BuiltinOr S.Expr
	}



compileDefns :: Scope
			 -> Map Name Defn
			 -> ErrorHandling (Map Name PolyValue)

compileDefns scope defns = do
	
	-- Explicit type signatures break dependency loops
	let explicitTypeSigs = Map.fromList [(name, ty)
		| (name, Defn {defnType = Just ty}) <- Map.toList defns]
	let namesWithTypeSigs = Map.keys explicitTypeSigs
	
	let scope' = scope {
		scopeValues = Map.map (\ (PolyFlavorType vars fty) -> pvalueWithoutImpl vars fty) explicitTypeSigs
			`Map.union` (scopeValues scope)
		}
	
	let clusters = map flattenSCC $ stronglyConnComp [let
		dependencies = case defnExpr defn of
			S.UserDefined e -> namesUsedInExpr e \\ namesWithTypeSigs
			S.Builtin _ -> []
		in ((name, defn), name, dependencies)
		| (name, defn) <- Map.toList defns]
		
	-- Because explicit type signatures break dependency loops, values with explicit type signatures
	-- will always appear alone in a one-element cluster. 
	
	pvalues <- compileClusters scope' clusters
		
	return pvalues
	
	where
		namesUsedInExpr (S.VarExpr _ name) = [name]
		namesUsedInExpr (S.ConExpr _ _) = []
		namesUsedInExpr (S.NumericLiteralExpr _ _) = []
		namesUsedInExpr (S.StringLiteralExpr _ _) = []
		namesUsedInExpr (S.LambdaExpr _ args body) = namesUsedInExpr body \\ map snd args
		namesUsedInExpr (S.ApplyExpr _ _ fun arg) = namesUsedInExpr fun ++ namesUsedInExpr arg
		namesUsedInExpr (S.CaseExpr _ subj pats) = namesUsedInExpr subj ++ concat [namesUsedInExpr exp \\ namesProducedInPattern pat | (pat,exp) <- pats]
		namesUsedInExpr (S.DoExpr _ stmts result) = namesUsedInStmts stmts
			where
				namesUsedInStmts ((S.BindStmt _ _ name value):rest) = namesUsedInExpr value ++ (namesUsedInStmts rest \\ [name])
				namesUsedInStmts ((S.ActStmt _ value):rest) = namesUsedInExpr value ++ namesUsedInStmts rest
				namesUsedInStmts [] = namesUsedInExpr result
		
		namesProducedInPattern (S.VarPattern _ name) = [name]
		namesProducedInPattern (S.ConApplyPattern _ _ pats) = concatMap namesProducedInPattern (map snd pats)
		namesProducedInPattern (S.WildcardPattern _) = []



compileClusters :: Scope
				-> [[(Name, Defn)]]
				-> ErrorHandling (Map Name PolyValue)

compileClusters = compileClusters' Map.empty
	where
		compileClusters' done scope [] = return done
		compileClusters' done scope (cluster:rest) = do
			clusterDone <- compileCluster scope cluster
			let scope' = scope { scopeValues = Map.union clusterDone (scopeValues scope) }
			compileClusters' (done `Map.union` clusterDone) scope' rest



compileCluster :: Scope
			   -> [(Name, Defn)]
			   -> ErrorHandling (Map Name PolyValue)

-- Handle a one-element cluster with explicit type signature
compileCluster scope [(name, Defn { defnType = Just givenPolyType, defnExpr = expr })] = do
	
	pvalue <- errorLoc ("In definition of '"++pretty name++"':") $
		compileExpr scope (Defn { defnType = Just givenPolyType, defnExpr = expr })
	
	return (Map.singleton name pvalue)

-- Handle a (possibly) multi-element cluster with no explicit type signatures
compileCluster scope cluster = do
	
	let (names, defns) = unzip cluster
	
	let nameStream = [Name (l:n) | n <- [""]++map show [1..], l <- ['a'..'z']]
	
	-- Construct a system of equalities that we must solve to determine the types of the
	-- definitions. This is done in the ConstraintMonad monad for convenience.
	(freeVars, constraints, (types, exprs')) <- runConstraintMonad nameStream $ do
		
		-- Make new variables to represent the types of each of the values in the cluster.
		assumedTypes <- sequence [do
			ty <- newFreeVar TypeKind ("type of '" ++ pretty name ++ "'")
			flavor <- newFreeVar monadKind ("flavor of '" ++ pretty name ++ "'")
			return (FlavorType flavor ty)
			| name <- names]
		
		-- Compute the types and the system of constraints on the assumption that the variables
		-- represent the actual types.
		let scopeWithAssumedTypes = scope {
			scopeValues = (Map.fromList
				[(name, pvalueWithoutImpl [] assumedType)
				| (name, assumedType) <- zip names assumedTypes])
				`Map.union` (scopeValues scope)
			}
		(computedTypes, exprs') <- liftM unzip $ sequence [let
			-- If a name appears in a cluster containing more than one name, it must be user-defined
			-- rather than built-in
			S.UserDefined exp = defnExpr defn
			in errorLocC ("In definition of '"++pretty name++"':") $
				desugarExpr scopeWithAssumedTypes exp
			| (name, defn) <- zip names defns]
		
		-- Make more constraints to bind the variables for the assumed types to the actual
		-- computed types.
		sequence [do
			constrainEqual assumedType computedType ("type of '" ++ pretty name ++ "'")
			constrainEqual assumedFlavor computedFlavor ("flavor of '" ++ pretty name ++ "'")
			| (name, FlavorType assumedFlavor assumedType, FlavorType computedFlavor computedType)
			<- zip3 names assumedTypes computedTypes]
		
		return (assumedTypes, exprs')
	
	-- Solve the system of constraints that we just made.
	(subs, stillFreeVars) <- solveConstraints freeVars constraints
	
	-- Now construct a PolyValue for each definition.
	pvalues <- sequence [errorLoc ("In definition of '"++pretty name ++ "'") $ do
		
		-- Substitute in the values that were found for variables
		let fty'@(FlavorType f' t') = substituteTypeInFlavorType subs fty
		let expr'' = substituteTypeInExpr subs expr'
		
		let driftingVars = filter
			(\fv -> not (
				varOccursInType (freeVarName fv) f' || varOccursInType (freeVarName fv) t'
				))
			stillFreeVars
		verifyNoDriftingVars driftingVars
		
		return (pvalueWithExprImpl [(freeVarName fv, freeVarKind fv) | fv <- stillFreeVars] fty' expr'')
		
		| (name, fty, expr') <- zip3 names types exprs'] 
	
	return (Map.fromList (zip names pvalues))



compileExpr :: Scope
			-> Defn
			-> ErrorHandling PolyValue

compileExpr scope (Defn { defnType = Nothing, defnExpr = S.UserDefined expr }) = do
	
	let nameStream = [Name (l:n) | n <- [""] ++ map show [1..], l <- ['a'..'z']]
	
	-- Construct a system of constraints that describes the types of the definitions
	(freeVars, constraints, (fty, expr')) <- runConstraintMonad nameStream (desugarExpr scope expr)
	
	-- Solve the system of constraints
	(subs, stillFreeVars) <- solveConstraints freeVars constraints
	
	-- Substitute the values we determined for the variables
	let expr'' = substituteTypeInExpr subs expr'
	let fty'@(FlavorType f' t') = substituteTypeInFlavorType subs fty
	
	let driftingVars = filter
		(\fv -> not (varOccursInType (freeVarName fv) f' || varOccursInType (freeVarName fv) t'))
		stillFreeVars
	verifyNoDriftingVars driftingVars
	
	return (pvalueWithExprImpl [(freeVarName fv, freeVarKind fv) | fv <- stillFreeVars] fty' expr'')

compileExpr scope (Defn { defnType = Just givenPolyType, defnExpr = S.UserDefined expr }) = do
	
	let (PolyFlavorType givenFreeVars (FlavorType givenFlavor givenType)) = givenPolyType
	
	let nameStream = [Name (l:n) | n <- [""]++map show [1..], l <- ['a'..'z']] \\ map fst givenFreeVars
	
	-- Construct a system of constraints
	(freeVars, constraints, expr') <- runConstraintMonad nameStream $ do
	
		(FlavorType computedFlavor computedType, expr') <- desugarExpr scope expr
		
		-- Constrain the computed type to match the specified type signature
		constrainEqual givenType computedType ("type of '" ++ pretty expr ++ "'")
		constrainEqual givenFlavor computedFlavor ("flavor of '" ++ pretty expr ++ "'")
		
		return expr'
	
	(subs, driftingVars) <- solveConstraints freeVars constraints
	
	-- All of the free vars from the original constraints should have been either given a value or
	-- mapped to one of the free vars from the given polymorphic type signature, so any remaining
	-- free vars after solving the constraints are drifting vars.

	verifyNoDriftingVars driftingVars

	-- Substitute in the solution that we found
	let expr'' = substituteTypeInExpr subs expr'
		
	return (pvalueWithExprImpl givenFreeVars (FlavorType givenFlavor givenType) expr'')

compileExpr scope (Defn { defnType = Nothing, defnExpr = S.Builtin _ } ) = do
	
	fail ("Built-in values must have an explicit type signature.")

compileExpr scope (Defn { defnType = Just ty, defnExpr = S.Builtin impl } ) = do
	
	let (PolyFlavorType freeVars fty) = ty
	
	return (PolyValue {
		pvalueFreeVars = freeVars,
		pvalueType = fty,
		pvalueImplExpr = Nothing,
		pvalueImplHaskell = Just impl
		})



pvalueWithoutImpl :: [(Name, Kind)] -> FlavorType -> PolyValue
pvalueWithoutImpl freeVars fty = PolyValue {
	pvalueFreeVars = freeVars,
	pvalueType = fty,
	pvalueImplExpr = Nothing,
	pvalueImplHaskell = Nothing
	}

pvalueWithExprImpl :: [(Name, Kind)] -> FlavorType -> Expr -> PolyValue
pvalueWithExprImpl freeVars fty expr = PolyValue {
	pvalueFreeVars = freeVars,
	pvalueType = fty,
	pvalueImplExpr = Just expr,
	pvalueImplHaskell = Just $ HaskellBackend.convertExpr expr
	}



verifyNoDriftingVars :: [FreeVar] -> ErrorHandling ()
verifyNoDriftingVars driftingVars =

	when (not (null driftingVars)) $
		fail ("The following variables are drifting:" ++
			concat ["\n" ++ pretty (freeVarName fv) ++ " :: " ++ pretty (freeVarKind fv) ++
				" (" ++ freeVarSrc fv ++ ")" | fv <- driftingVars])
