module DesugarType where



import Control.Monad
import Data.Map (Map)
import qualified Data.Map as Map

import Core
import Scope
import qualified Syntax as S
import Pretty
import Names
import ErrorHandling


{-
desugarType converts a Surface.Type to a Core.Type. It also computes its kind.
-}



-- The local type scope contains the kinds of local type variables. For example, in desugaring the
-- type signature "id :: (a :: *, f :: !) => [f] (a -> a)", the variables 'a' and 'f' would be in the
-- local type scope.

type LocalTypeScope = Map Name Kind



desugarType :: (Scope, LocalTypeScope, Maybe Type) -> S.Type -> ErrorHandling (Kind, Type)

desugarType (scope, localScope, defFlavor) (S.VarType span v) =
	case Map.lookup v localScope of
		-- Determine the variable's kind from the local scope
		Just k -> return (k, VarType v k)
		Nothing -> errorLoc ("At "++pretty span++":") $ fail ("Type-variable '"++pretty v++"' not in scope.")

desugarType (scope, localScope, defFlavor) (S.ConType span c) =
	case Map.lookup c (scopeTypes scope) of
		Just ti -> return (typeInfoKind ti, ConType c (typeInfoKind ti))
		Nothing -> errorLoc ("At "++pretty span++":") $ fail ("Type-constructor '"++pretty c++"' not in scope.")

desugarType scopeAndMore (S.ApplyType span fun arg) =
	errorLoc ("In "++pretty span++":") $ do
	
		(funKind, fun') <- desugarType scopeAndMore fun
		
		case funKind of
		
			FunctionKind argKind retKind -> do
				(inferred, arg') <- desugarType scopeAndMore arg
				verifyKind inferred argKind ("the argument to '"++pretty fun++"'")
				return (retKind, ApplyType fun' arg')
			
			_ -> fail ("Cannot apply type '"++pretty fun++"', which has kind '"++pretty funKind++"'.")

desugarType scopeAndMore@(scope, localScope, defFlavor) function@(S.FunctionType span arg maybeFlavor ret) =
	errorLoc ("In "++pretty span++":") $ do
		
		arg' <- errorLoc ("In argument type:") $
			desugarFieldType scopeAndMore arg
		
		flavor' <- case maybeFlavor of
			Just f -> do
				(fk, f') <- desugarType scopeAndMore f
				verifyKind fk monadKind ("the function's flavor")
				return f'
			Nothing -> case defFlavor of
				Just f -> return f
				Nothing -> fail ("You must explicitly specify a flavor for ("++pretty arg++" -> "++
					pretty ret++") because there is no default flavor in this context. To \
					\explicitly specify a flavor, write ("++pretty arg++" -> [some flavor] " ++
					pretty ret++").")
		
		(rk, ret')  <- errorLoc ("In return type:") $
			desugarType scopeAndMore ret
		verifyKind rk TypeKind ("the function's return type")
		
		return (TypeKind, FunctionType arg' (FlavorType flavor' ret'))



desugarFieldType :: (Scope, LocalTypeScope, Maybe Type) -> S.FieldType -> ErrorHandling FieldType

desugarFieldType scopeAndMore (S.StrictField t) = do
	(tk, t') <- desugarType scopeAndMore t
	-- Probably the least descriptive error message ever... oh well, errorLoc information should
	-- clue them toward what is going wrong
	verifyKind tk TypeKind ("the type")
	return (StrictField t')

desugarFieldType scopeAndMore@(_, _, defaultFlavor) (S.LazyField maybeFlavor t) = do
	(tk, t') <- desugarType scopeAndMore t
	verifyKind tk TypeKind ("the type")
	f' <- case maybeFlavor of
		Just f -> do
			(fk, f') <- desugarType scopeAndMore f
			verifyKind fk monadKind ("the flavor")
			return f'
		Nothing -> case defaultFlavor of
			Just f -> return f
			Nothing -> fail ("You must explicitly specify a flavor for {"++pretty t++"} because \
				\there is no default flavor in this context. To explicitly specify a flavor, \
				\write {[some flavor] "++pretty t++"}.")
	return (LazyField f' t')



verifyKind :: Kind -> Kind -> String -> ErrorHandling ()
verifyKind k1 k2 what =
	when (k1 /= k2) (fail ("The kind of "++what++" is '"++pretty k1++"', but it should be '"++pretty k2++"'."))
