module HaskellBackend where

import Core
import Names
import Pretty
import Scope

import Control.Monad.Writer
import Data.Char
import Data.List
import Data.Map (Map)
import qualified Data.Map as Map



type HSDecl = String
type HSExpr = String
type HSType = String
type HSPattern = String



-- The 'prefix' argument is used for name mangling, to avoid conflicts between user-defined names
-- and internal or Haskell-defined names. The prefixes used are as follows:
-- "v" for user-defined global variables
-- "s" for strict user-defined local variables
-- "l" for lazy user-defined local variables
-- "T" for user-defined global types
-- "t" for user-defined and internally-generated type variables
-- "C" for user-defined constructors
-- "c" for the internally-generated flavored version of a user-defined constructor
convertName :: String -> Name -> String

convertName prefix (Name x) = prefix ++ "_" ++ x



-- Not used in practice
convertKind :: Kind -> String

convertKind TypeKind = "*"

convertKind (FunctionKind a r) = "(" ++ convertKind a ++ " -> " ++ convertKind r ++ ")"



convertType :: Type -> HSType

convertType (VarType n _) =
    convertName "t" n

convertType (ConType n _) =
    convertName "T" n

convertType (ApplyType a b) =
    "(" ++ convertType a ++ " " ++ convertType b ++ ")"

convertType (FunctionType a r) =
    "(" ++ convertFieldType a ++ " -> " ++ convertFlavorType r ++ ")"

convertFieldType :: FieldType -> HSType

convertFieldType (StrictField t) =
    convertType t

convertFieldType (LazyField f t) =
    convertFlavorType (FlavorType f t)

convertFlavorType :: FlavorType -> HSType

convertFlavorType (FlavorType f t) =
    "(" ++ convertType f ++ " " ++ convertType t ++ ")"



convertExpr :: Expr -> HSExpr
convertExpr = convertExpr' Map.empty

convertExpr' :: Map Name Strictness -> Expr -> HSExpr

convertExpr' scope (VarExpr (name, fty)) =
    case Map.lookup name scope of
        Just Strict -> "(return " ++ convertName "s" name ++ ")" 
        Just Lazy -> convertName "l" name
        Nothing -> convertName "v" name

convertExpr' scope (ConExpr (name, fty)) =
    -- Use the lower-case 'c' name to get the automatically-generated flavored constructor instead
    -- of the constructor itself
    convertName "c" name

convertExpr' scope (NumericLiteralExpr value flavor) =
    "(return (" ++ show value ++ "))"   -- Extra parentheses in case of negatives

convertExpr' scope (StringLiteralExpr value flavor) =
    "(return " ++ show value ++ ")"

convertExpr' scope (LambdaExpr name (StrictField _) body) =
    "(return (\\ (!" ++ convertName "s" name ++ ") -> " ++ convertExpr' scope' body ++ "))"
    where scope' = Map.insert name Strict scope

convertExpr' scope (LambdaExpr name (LazyField _ _) body) =
    "(return (\\" ++ convertName "l" name ++ " -> " ++ convertExpr' scope' body ++ "))"
    where scope' = Map.insert name Lazy scope

convertExpr' scope (ApplyExpr Strict fun arg) =
    "(" ++ convertExpr' scope fun ++ " $$! " ++ convertExpr' scope arg ++ ")"

convertExpr' scope (ApplyExpr Lazy fun arg) =
    "(" ++ convertExpr' scope fun ++ " $$ " ++ convertExpr' scope arg ++ ")"

convertExpr' scope (CaseExpr subject branches) =
    "(do { s <- " ++ convertExpr' scope subject ++ "; case s of { " ++
        concat [let
            (pat', names) = runWriter (convertPattern' Strict pat)
            scope' = Map.fromList names `Map.union` scope
            in pat' ++ " -> " ++ convertExpr' scope' branch ++ "; "
            | (pat, branch) <- branches] ++
        "}; })"

convertExpr' scope (DoExpr stmts result) =
    "(do { " ++ convertStmts scope stmts result ++ "; })"
    where
        convertStmts scope [] result = convertExpr' scope result
        
        convertStmts scope ((BindStmt name (StrictField _) value) : rest) result =
            "(!" ++ convertName "s" name ++ ") <- " ++ convertExpr' scope value ++ "; " ++
            convertStmts (Map.insert name Strict scope) rest result
            
        convertStmts scope ((BindStmt name (LazyField _ _) value) : rest) result =
            "let {" ++ convertName "l" name ++ " = " ++ convertExpr' scope value ++ "}; " ++
            convertStmts (Map.insert name Lazy scope) rest result
            
        convertStmts scope ((ActStmt value) : rest) result =
            convertExpr' scope value ++ "; " ++
            convertStmts scope rest result



convertPolyValue :: PolyValue -> HSExpr

convertPolyValue pv =
    case pvalueImplHaskell pv of
        Just str -> str
        Nothing -> error ("HaskellBackend.convertPolyValue: No suitable implementation.")



convertPattern' :: Strictness -> Pattern -> Writer [(Name, Strictness)] HSPattern

convertPattern' Strict (VarPattern (name, StrictField _)) = do
    tell [(name, Strict)]
    return (convertName "s" name)

convertPattern' Lazy (VarPattern (name, LazyField _ _)) = do
    tell [(name, Lazy)]
    return (convertName "l" name)

convertPattern' _ (VarPattern _) =
    error "Strictness on var doesn't match strictness on field."

convertPattern' Strict (ConApplyPattern (con, _) fields) = do
    fields' <- mapM (uncurry convertPattern') fields
    return ("(" ++ convertName "C" con ++ concat [" " ++ f | f <- fields'] ++ ")")

convertPattern' Lazy (ConApplyPattern _ _) = do
    error "I lied, we don't actually support pattern matching on lazy fields."

convertPattern' _ (WildcardPattern) =
    return "_"



runtime = "{-# LANGUAGE KindSignatures, BangPatterns #-}\n\
\\n\
\import Control.Monad\n\
\import IO\n\
\\n\
\-- Monad-embed runtime support\n\
\\n\
\-- Application for strict functions\n\
\($$!) :: Monad m => m (a -> m b) -> m a -> m b\n\
\f $$! x = do\n\
\    f' <- f\n\
\    x' <- x\n\
\    f' x'\n\
\\n\
\-- Application for lazy functions\n\
\($$) :: (Monad m, Monad n) => m (n a -> m b) -> n a -> m b\n\
\f $$ x = do\n\
\    f' <- f\n\
\    f' x\n\
\\n\
\convertBool :: Bool -> T_Bool\n\
\convertBool True = C_True\n\
\convertBool False = C_False\n\
\\n\
\data ComposeType trans m x = ComposeType (m (trans m x))\n\
\\n\
\class Transformer t where\n\
\    returnT :: Monad f => f (a -> f (t f a))\n\
\    bindT :: Monad f => f (t f a -> f ((a -> f (t f b)) -> f (t f b)))\n\
\\n\
\instance (Transformer trans, Monad f) => Monad (ComposeType trans f) where\n\
\    return x = ComposeType $ do\n\
\        r <- returnT\n\
\        r x\n\
\    (>>=) (ComposeType x) f = ComposeType $ do\n\
\        b <- bindT\n\
\        x' <- x\n\
\        b' <- b x'\n\
\        b' (\\y -> case f y of ComposeType y' -> y')\n\
\\n\
\main = v_main\n\
\\n\
\-- Actual definitions\n\
\\n"

convertProgram :: Scope -> String
convertProgram scope =
    runtime ++
    concat [decl ++ "\n" | decl <- convertDecls scope]



convertDecls :: Scope -> [HSDecl]
convertDecls scope =
    map (uncurry (convertValueDecl scope)) (Map.toList (scopeValues scope)) ++
    map (uncurry (convertTypeDecl scope)) (Map.toList (scopeTypes scope)) ++
    map (uncurry (convertTransformerDecl scope)) (Map.toList (scopeTransformers scope))



convertValueDecl :: Scope -> Name -> PolyValue -> HSDecl
convertValueDecl scope name pvalue =
    convertName "v" name ++ " :: " ++
        (if null constraints then "" else "(" ++ intercalate ", " constraints ++ ") => ") ++
        convertFlavorType (pvalueType pvalue) ++ "\n" ++
    convertName "v" name ++ " = " ++ convertPolyValue pvalue ++ "\n"
    where constraints = makeConstraints (pvalueFreeVars pvalue)



convertTransformerDecl :: Scope -> Name -> TransformerInfo -> HSDecl
convertTransformerDecl scope name info =
    "instance (" ++ intercalate ", " constraints ++ ") => Transformer (" ++ transformer ++ ") where\n\
    \    returnT = " ++ convertPolyValue (transformerReturn info) ++ "\n\
    \    bindT = " ++ convertPolyValue (transformerBind info) ++ "\n"
    where
        conKind = case Map.lookup name (scopeTypes scope) of
            Just ti -> typeInfoKind ti
            Nothing -> error "HaskellBackend.convertMonad: Missing kind info for a type"
        fieldsOfKind TypeKind = []
        fieldsOfKind (FunctionKind f k) = f : fieldsOfKind k
        fieldKinds = take (length (fieldsOfKind conKind) - 2) (fieldsOfKind conKind)
        fieldNames = [Name ("f" ++ show n) | n <- [1 .. length fieldKinds]]
        transformer = convertName "T" name ++ concat [" " ++ convertName "t" n | n <- fieldNames]
        constraints = makeConstraints (zip fieldNames fieldKinds)



convertTypeDecl :: Scope -> Name -> TypeInfo -> HSDecl
convertTypeDecl scope name (info @ AlgebraicTypeInfo {}) =
    "data " ++
        convertName "T" name ++
        concat [" (" ++ convertName "t" name ++ " :: " ++ convertKind kind ++ ")"
            | (name, kind) <- typeInfoFreeVars info] ++
        " = " ++
        intercalate " | " [
            convertName "C" name ++ concat [ let
                bang = case ty of
                    StrictField _ -> "!"
                    LazyField _ _ -> ""
                in " " ++ bang ++ convertFieldType ty
                | ty <- conFieldTypes con]
            | name <- typeInfoCons info, let Just con = Map.lookup name (scopeCons scope)] ++
        "\n" ++
    concat [convertName "c" name ++ " = " ++
        foldr (\fieldName ret -> "return (\\" ++ fieldName ++ " -> " ++ ret ++ ")")
            ("return (" ++ convertName "C" name ++ concat [" " ++ n | n <- fieldNames] ++ ")")
            fieldNames ++ "\n" ++
        convertName "c" name ++ " :: (Monad " ++ convertName "t" (Name "f_") ++ ") => " ++
            (convertFlavorType (conTypeInFlavor con (VarType (Name "f_") monadKind))) ++ "\n"
        | name <- typeInfoCons info
        , let Just con = Map.lookup name (scopeCons scope)
        , let fieldNames = ["f" ++ show i | i <- [1..length (conFieldTypes con)]]]
convertTypeDecl scope name (info @ BuiltinTypeInfo {}) =
    "type " ++
        convertName "T" name ++
        " = " ++
        (case typeInfoImplHaskell info of
            Just s -> s
            Nothing -> error ("HaskellBackend.convertTypeDecl: '" ++ pretty name ++ "' lacks a \
                \suitable implementation.")) ++
        "\n"


-- The semantics of Core say that instances should not be checked until runtime. The way we simulate
-- this in Haskell is by constraining everything of kind (* -> *) to be a Monad and everything of
-- kind ((* -> *) -> * -> *) to be a Transformer, and then having "default" instances of Monad and
-- Transformer that raise a runtime error.
makeConstraints :: [(Name, Kind)] -> [String]
makeConstraints = concatMap (uncurry makeConstraintsForOne)
    where
        makeConstraintsForOne name TypeKind =
            []
        makeConstraintsForOne name (TypeKind `FunctionKind` TypeKind) =
            ["Monad " ++ convertName "t" name]
        makeConstraintsForOne name ((TypeKind `FunctionKind` TypeKind) `FunctionKind` TypeKind `FunctionKind` TypeKind) =
            ["Transformer " ++ convertName "t" name]
        -- TODO: If we have a free variable like "f :: * -> * -> *", then we should have a
        -- constraint that says "Monad (f a)".


