module TypeInference where



import Control.Monad 
import Data.Map (Map)
import qualified Data.Map as Map

-- For debugging
import Debug.Trace

import Core
import ErrorHandling
import Names
import Pretty
import Substitute
import qualified Syntax as S



data FreeVar = FreeVar {
	freeVarName :: Name,
	freeVarKind :: Kind,
	freeVarSrc :: String
	}

data Constraint = Constraint {
	constraintValues :: (Type, Type),
	constraintOriginalValues :: (Type, Type),
	constraintSrc :: String
	}



{-
The constraint monad is a monad used for compiling expressions. It has three roles:

1. Error handling
2. Generating unique names
3. Keeping track of the constraints generated (much like Writer)

Perhaps it would be better if it were done using monad transformers.
-}

data ConstraintMonad a =
	ConstraintMonad ([Name] ->
		ErrorHandling ([Name], [FreeVar], [Constraint], a)
		)
instance Monad ConstraintMonad where
	return x = ConstraintMonad (\names -> return (names, [], [], x))
	ConstraintMonad a >>= fb = ConstraintMonad
		(\names -> do
			(names', fvs1, cnts1, x) <- a names
			(names'', fvs2, cnts2, y) <- case fb x of ConstraintMonad b -> b names'
			return (names'', fvs1++fvs2, cnts1++cnts2, y)
			)
	fail x = ConstraintMonad (const (fail x))



-- Generate a new free variable of the given kind, and record its existence in the constraint monad.
-- Then return the variable, in Type form.
newFreeVar :: Kind -> String -> ConstraintMonad Type
newFreeVar kind src =
	ConstraintMonad (\(name:names) -> do
		loc <- getErrorLoc
		let fv = FreeVar name kind (loc ++ src)
		return (names, [fv], [], VarType name kind)
		)

-- Record an equality constraint in the constraint monad.
constrainEqual :: Type -> Type -> String -> ConstraintMonad ()
constrainEqual t1 t2 src =
	if typeKind t1 == typeKind t2
	then ConstraintMonad (\names -> do
		loc <- getErrorLoc
		let c = Constraint (t1, t2) (t1, t2) (loc ++ "When computing " ++ src ++ ":")
		return (names, [], [c], ()))
	else error ("Internal error: in constraint '"++src++"' kind of '"++pretty t1++"' does not \
		\match kind of '"++pretty t2++"'.")

runConstraintMonad :: [Name] -> ConstraintMonad a -> ErrorHandling ([FreeVar], [Constraint], a)
runConstraintMonad names (ConstraintMonad f) = case runErrorHandling (f names) of
	Left err -> fail err
	Right (names', freeVars, constraints, x) -> return (freeVars, constraints, x)

errorLocC :: String -> ConstraintMonad x -> ConstraintMonad x
errorLocC loc (ConstraintMonad f) = ConstraintMonad (\names -> errorLoc loc (f names))



data InferenceError = InferenceError String String (Type, Type)



solveConstraints :: [FreeVar] -> [Constraint] -> ErrorHandling (Map Name Type, [FreeVar])

solveConstraints freeVars constraints = perhapsTrace $

	case solveConstraints' (Map.empty, freeVars, []) constraints of
	
		(subs, stillFreeVars, []) -> return (subs, stillFreeVars)
		(_, _, errors) -> fail $ case head errors of
			InferenceError src reason (a, b) ->
				src ++ "\nCannot solve equality between:\n" ++ pretty a ++ "\n" ++
				pretty b ++ "\n" ++ reason ++ "."
	
	where
		perhapsTrace = id
		{-
		perhapsTrace = Debug.Trace.trace (
			"CONSTRAINTS:\n\n" ++ concat [
			constraintSrc c ++ "\n" ++
			pretty (fst (constraintOriginalValues c)) ++ "\n" ++
			pretty (snd (constraintOriginalValues c)) ++ "\n" ++
			"\n"
			| c <- constraints] ++
			"FREE VARS:\n\n" ++ concat [
			freeVarSrc v ++ "\n" ++
			pretty (freeVarName v) ++ " :: " ++ pretty (freeVarKind v) ++ "\n" ++
			"\n"
			| v <- freeVars]
			)
		-}



-- We do not bail out immediately when we detect an error. Instead, we record it as an
-- InferenceError and then keep trying to solve constraints. This lets us produce multiple error
-- messages if there are multiple errors, and also to use information from the later constraints to
-- make earlier error messages more specific.

solveConstraints' :: (Map Name Type, [FreeVar], [InferenceError])
				 -> [Constraint]
				 -> (Map Name Type, [FreeVar], [InferenceError])

solveConstraints' (done, freeVars, errors) [] = (done, freeVars, errors)

solveConstraints' (done, freeVars, errors) (c : rest) =
	
	case constraintValues c of
		
		(a, b) | typeKind a /= typeKind b -> error ("INTERNAL ERROR:\n"++constraintSrc c++"\nBad \
			\constraint: '"++pretty a++"' has kind '"++pretty (typeKind a)++"', but '"++pretty b++
			"' has kind '"++pretty (typeKind b)++"'. Original equality is between '"++
			pretty (fst (constraintOriginalValues c))++"' and '"++
			pretty (snd (constraintOriginalValues c))++".")
		
		(VarType name1 _, VarType name2 _) | name1 == name2 ->
			solveConstraints' (done, freeVars, errors) rest
		
		(VarType name kind, x) | name `elem` map freeVarName freeVars && varOccursInType name x ->
			
			-- When an error is detected, record it and then try to solve the rest of the
			-- constraints anyway.
			solveConstraints' (done, freeVars, e : errors) rest
			
			where e = InferenceError (constraintSrc c) "Occurs check failed"
				(constraintOriginalValues c)
		
		(VarType name kind, x) | name `elem` map freeVarName freeVars -> let
			
			sub = substituteTypeInType (Map.singleton name x)
			
			-- Now that we have a value for the variable, remove it from the free var list
			freeVars' = filter ((/= name) . freeVarName) freeVars
			
			-- Substitute in the new value wherever the variable appeared
			done' = Map.insert name x (Map.map sub done)
			rest' = [c { constraintValues = (sub v1, sub v2), constraintOriginalValues = (sub v3, sub v4) }
				| c <- rest, let (v1,v2) = constraintValues c, let (v3,v4) = constraintOriginalValues c]
			errors' = [InferenceError src reason (sub a, sub b)
				| InferenceError src reason (a, b) <- errors]
			
			in solveConstraints' (done', freeVars', errors') rest'
		
		(x, VarType name kind) | name `elem` map freeVarName freeVars ->
			-- Swap the order and then put it back onto the stack, so that it is handled by the
			-- cases immediately above us. This way we don't have to duplicate the code.
			solveConstraints' (done, freeVars, errors)
				(c { constraintValues = (VarType name kind, x) } : rest)
		
		(ConType c1 _, ConType c2 _) | c1 == c2 ->
			solveConstraints' (done, freeVars, errors) rest

		(ApplyType fun1 arg1, ApplyType fun2 arg2) ->
			solveConstraints' (done, freeVars, errors)
				(c { constraintValues = (fun1, fun2) } :
				 c { constraintValues = (arg1, arg2) } :
				 rest)
		
		(FunctionType (StrictField arg1) (FlavorType flavor1 ret1),
		 FunctionType (StrictField arg2) (FlavorType flavor2 ret2)) ->
			solveConstraints' (done, freeVars, errors)
				(c { constraintValues = (arg1, arg2) } :
				 c { constraintValues = (ret1, ret2) } :
				 c { constraintValues = (flavor1, flavor2) } :
				 rest)
		
		(FunctionType (LazyField argFlavor1 arg1) (FlavorType flavor1 ret1),
		 FunctionType (LazyField argFlavor2 arg2) (FlavorType flavor2 ret2)) ->
			solveConstraints' (done, freeVars, errors)
				(c { constraintValues = (argFlavor1, argFlavor2) } :
				 c { constraintValues = (arg1, arg2) } :
				 c { constraintValues = (ret1, ret2) } :
				 c { constraintValues = (flavor1, flavor2) } :
				 rest)
		
		(x, y) ->
			solveConstraints' (done, freeVars, e : errors) rest
			where e = InferenceError (constraintSrc c) "No match" (constraintOriginalValues c)



varOccursInType :: Name -> Type -> Bool
varOccursInType name (VarType name' _) =
	name == name'
varOccursInType name (ConType _ _) =
	False
varOccursInType name (ApplyType fun arg) =
	varOccursInType name fun || varOccursInType name arg
varOccursInType name (FunctionType (StrictField arg) (FlavorType flavor ret)) =
	varOccursInType name arg || varOccursInType name flavor || varOccursInType name ret
varOccursInType name (FunctionType (LazyField argFlavor arg) (FlavorType flavor ret)) =
	varOccursInType name argFlavor || varOccursInType name arg || varOccursInType name flavor || varOccursInType name ret

